unit
  UVDP;

interface

uses
   mzio;

  const
    aVDPInitSMS: array[0..15] of byte =
     ($00,$60,$0E,$00,$00,$7F,$00,$00,
      $00,$00,$FF,$00,$00,$00,$00,$00);
    aVDPInitColeco: array[0..15] of byte =
     ($00,$10,$00,$00,$00,$00,$00,$00,
      $00,$00,$00,$00,$00,$00,$00,$00);
    aMSK: array[0..3, 0..3] of LongInt =
     (($7F,$00,$3F,$00),  (* SCR 0:TEXT 40x24  *)
      ($7F,$FF,$3F,$FF),  (* SCR 1:TEXT 32x24  *)
      ($7F,$80,$3C,$FF),  (* SCR 2:BLK 256x192 *)
      ($7F,$00,$3F,$FF)); (* SCR 3:64x48x16    *)


  type
    TEmulObject = object
      constructor Init(P: Pointer; N: Integer);
      destructor  Done; virtual;
    end;


    (*
    memory = array[0..$FFFF] of Byte;
    PByteArray = ^memory;
    *)

    TVDP=object(TEmulObject)
      nRAddr, nWAddr: Word;
      aVDP: array[0..15] of Byte;
      nCurLine, nMinLine, nMaxLine: Integer;
      nColTab,
      nChrTab,
      nSprTab,
      nSprGen,
      nChrGen,
      nChrPal,
      nSprPal: Word;
      nVKey, nVR,
      nVDPStatus,
      nScrMode,
      nFGColor,
      nBGColor: Byte;
      nScrollX, nScrollY: Byte;
      constructor Init(P: Pointer; N: Integer);
      procedure   OutByte(nAddr, nValue: Word);
      procedure   SetScroll;
      function    Interrupt: LongInt;
      function    ReadVRAM: Byte;
      procedure   WriteVRAM(nValue: Byte);
      procedure   Sprites;
      destructor  Done; virtual;
    end;

implementation

constructor TEmulObject.Init; begin end;
destructor TEmulObject.Done; begin end;


constructor TVDP.Init;
var
  I: Integer;
begin
  for i := 0 to N-1 do
    aVDP[i] := Byte(PChar(P)[i]);
  nVKey := 1;
  nVDPStatus := $9F;
  nFGColor := 0;
  nBGColor := 0;
  nScrMode := 0;
  nChrTab := 0;
  nColTab := 0;
  nChrGen := 0;
  nSprTab := 0;
  nSprGen := 0;
  nChrTab := ((LongInt(aVDP[2]) and aMSK[nScrMode, 0]) shl 10);
  nChrGen := ((LongInt(aVDP[4]) and aMSK[nScrMode, 2]) shl 11);
  nColTab := ((LongInt(aVDP[3]) and aMSK[nScrMode, 1]) shl 6);
  nSprTab := ((LongInt(aVDP[5]) and aMSK[nScrMode, 3]) shl 7);
  nSprGen := (LongInt(aVDP[6]) shl 11);
  nWAddr := 0;
end;

procedure TVDP.OutByte;
var
  J : LongInt;
begin
  case nAddr of
     0: begin
       case (((nValue and $0E) shr 1) or (aVDP[1] and $18)) of
         $10: J := 0;
         $00: J := 1;
         $01: J := 2;
         $08: J := 3;
         else J := nScrMode;
       end;
       if (J <> nScrMode) then begin
         nChrTab := ((LongInt(aVDP[2]) and aMSK[J, 0]) shl 10);
         nChrGen := ((LongInt(aVDP[4]) and aMSK[J, 2]) shl 11);
         nColTab := ((LongInt(aVDP[3]) and aMSK[J, 1]) shl 6);
         nSprTab := ((LongInt(aVDP[5]) and aMSK[J, 3]) shl 7);
         nSprGen := (LongInt(aVDP[6]) shl 11);
         nScrMode := J;
         { Iniciar Cores e Modo Grafico }
       end
     end;
     1: begin
       case (((aVDP[0] and $0E) shr 1) or (nValue and $18)) of
         $10: J := 0;
         $00: J := 1;
         $01: J := 2;
         $08: J := 3;
         else J := nScrMode;
       end;
       if (J <> nScrMode) then begin
         nChrTab := ((LongInt(aVDP[2]) and aMSK[J, 0]) shl 10);
         nChrGen := ((LongInt(aVDP[4]) and aMSK[J, 2]) shl 11);
         nColTab := ((LongInt(aVDP[3]) and aMSK[J, 1]) shl 6);
         nSprTab := ((LongInt(aVDP[5]) and aMSK[J, 3]) shl 7);
         nSprGen := (LongInt(aVDP[6]) shl 11);
         nScrMode := J;
         { Iniciar Cores e Modo Grafico }
       end;
     end;
     2: nChrTab := ((LongInt(nValue) and aMSK[nScrMode, 0]) shl 10);
     3: nColTab := ((LongInt(nValue) and aMSK[nScrMode, 1]) shl 6);
     4: nChrGen := ((LongInt(nValue) and aMSK[nScrMode, 2]) shl 11);
     5: nSprTab := ((LongInt(nValue) and aMSK[nScrMode, 3]) shl 7);
     6: begin
       nValue := nValue and $3F;
       nSprGen := (LongInt(nValue) shl 11);
     end;
     7: begin
       nFGColor := nValue shr 4;
       nBGColor := nValue and $0F;
       { Inializar Cores }
     end;
  end;
  aVDP[nAddr] := nValue;
end;


procedure TVDP.SetScroll;
begin
  if ((aVDP[0] and $40)<>0) and (nCurLine<(nMinLine+16)) then
    nScrollX := 0
  else
    nScrollX := 256-aVDP[8];
  nScrollY := aVDP[9];
end;

procedure TVDP.Sprites;
var
  nLin, nCol, nPixelLine, Y, nSpriteTable, C, nSpriteCol, N, S: Integer;
begin
  S := nSprTab;
  N :=0;
  while (N<32) and (rom^[S]<>208) do begin
    Inc(N);
    S := S+4;
  end;

  if ((aVDP[1] and $02)<>0) then begin
    Dec(S, 4);
    while (N<>0) do begin

      if ((rom^[S+3]) and $80)<>0 then
       nSpriteCol := rom^[S+1]-32
      else
       nSpriteCol := rom^[S+1];
      C := rom^[S+3] and $0F;

      if ((nSpriteCol<=(256-16)) and (nSpriteCol>=0) and (C<>0)) then begin
        nSpriteTable := nSprGen+((rom^[S+2] and $FC) shl 3);
        C := C+10;
        Y := rom^[S+0]+1;

        nLin := 0;
        if (Y<192) then begin
          nLin := Y;
          if Y>176 then Y := 192-Y
          else Y := 16;
        end
        else begin
          nSpriteTable := nSpriteTable+256-Y;
          if Y>240 then Y := Y-240
          else Y := 0;
        end;

        while Y<>0 do begin
          nPixelLine := rom^[nSpriteTable+0]*256+rom^[nSpriteTable+16];
          for nCol := 0 to 15 do begin
            if (nPixelLine and $8000)<>0 then Setpixel(nSpriteCol+nCol, nLin, C);
            nPixelLine := nPixelLine shl 1;
          end;

          Inc(nLin);
          Dec(Y);
          Inc(nSpriteTable);
        end;
      end;
      Dec(N);
      Dec(S, 4);
    end;
  end
  else begin
    Dec(S, 4);
    while N<>0 do begin
      if (rom^[S+3] and $80)<>0 then
        nSpriteCol := rom^[S+1]-32
      else
        nSpriteCol := rom^[S+1];
      C := rom^[S+3]and $0F;
      if ((nSpriteCol<=(256-8)) and (nSpriteCol>=0) and (C<>0)) then begin
        nSpriteTable := nSprGen+(rom^[S+2] shl 3);
        C := C+10;
        Y := rom^[S+0]+1;

        nLin := 0;
        if (Y<192) then begin
          nLin := Y;
          if (Y>184) then Y := 192-Y
          else Y := 8;
        end
        else begin
         nSpriteTable := nSpriteTable+256-Y;
         if (Y>248) then Y := Y-248
         else Y := 0;
        end;
        while Y<>0 do begin
          nPixelLine := rom^[nSpriteTable];

          for nCol := 0 to 7 do begin
            if (nPixelLine and $80)<>0 then Setpixel(nSpriteCol+nCol, nLin, C);

            nPixelLine := nPixelLine shl 1;
          end;
          Dec(Y);
          Inc(nSpriteTable);
          Inc(nLin);
        end;
      end;
      Dec(N);
      Dec(S, 4);
    end; {while N<>0}

  end; {of else}

end;

function TVDP.Interrupt: LongInt;
var
  FC,BC: Byte;
  nResult, C, PGT, CLT, K, I, J, X, Y, T, S, P, N: Integer;

const
  WIDTH=256;
  HEIGHT=256;
begin


  if (nVKey<> 0)
   then
  BEGIN

  case nScrMode of
    0: begin
      T := nChrTab;
      FC := nFGColor;
      BC := nBGColor;
      for Y := 23 downto 0 do begin
        for x := 0 to 39 do begin
          S := nChrGen+(rom^[T] shl 3);
          for i := 7 downto 0 do begin
            K := rom^[S];
            for j := 0 to 5 do begin
              if (K and 128)<>0 then
                Setpixel(6*X+j, 8*Y+i, FC+10)
              else
                Setpixel(6*X+j, 8*Y+i, BC+10);
              k := k shl 1;
            end;
            Inc(S);
          end;
          Inc(T);
        end;
      end;
    end;

    1: begin
      T := nChrTab;
      for Y := 23 downto 0 do begin
        for x := 0 to 31 do begin
          S := nChrGen+(rom^[T] shl 3);
          BC := rom^[nColTab+rom^[T] shr 3];
          FC := BC shr 4;
          BC := BC and $0F;

          for i := 7 downto 0 do begin
            K := rom^[S];
            for j := 0 to 7 do begin
              if (K and 128)<>0 then
                Setpixel(8*X+j, 191-(8*Y+i), FC+10)
              else
                Setpixel(8*X+j, 191-(8*Y+i), BC+10);
              k := k shl 1;
            end;
            Inc(S);



          end;




          Inc(T);
        end;
      end;
      Sprites;
    end;

    2: begin
      PGT := nChrGen;
      CLT := nColTab;
      T := nChrTab;

      for n := 0 to 2 do begin
        for y := 0 to 7 do begin
          for x := 0 to 31 do begin
            S := PGT+((rom^[T]) shl 3);
            C := CLT+((rom^[T]) shl 3);

            for i := 0 to 7 do begin
              K := (rom^[S]);
              BC := rom^[C];
              FC := BC shr 4;
              BC := BC and $0F;

              for j := 0 to 7 do begin
                if (K and 1)<>0 then
                  Setpixel(8*X+7-j, n*64+8*Y+i, FC+10)
                else
                  Setpixel(8*X+7-j, n*64+8*Y+i, BC+10);

                k := k shr 1;

              end;

              C := C+1;
              S := S+1;
            end;
            T := T+1;
          end;
        end;
        PGT := PGT+$800;
        CLT := CLT+$800;
      end;
      Sprites;
    end;
  end;


  nVDPStatus := (nVDPStatus and $DF) or $80;

  if (aVDP[1] and $20)=0 then
    Interrupt := $00000066
  else
    Interrupt :=$01000000;
  Interrupt := $66;

  END
  else
     Interrupt := $00000000;;

end;


function TVDP.ReadVRAM: Byte;
begin
  ReadVRAM := rom^[nRAddr];
  nRAddr := (nRAddr+1) and $3FFF;
end;

procedure TVDP.WriteVRAM(nValue: Byte);
begin
  rom^[nWAddr] := nValue;
  nWAddr := (nWAddr+1) and $3FFF;
end;


destructor TVDP.Done;
begin

end;


end.